import logging
import typing
from typing import Dict
from typing import Iterable
from typing import Optional
from typing import Type
from typing import Union

from werkzeug import Request
from werkzeug import Response

from clarin.sru.constants import SRUVersion
from clarin.sru.exception import SRUConfigException
from clarin.sru.queryparser import SRUQueryParserRegistry
from clarin.sru.server.auth import SRUAuthenticationInfoProvider
from clarin.sru.server.auth import SRUAuthenticationInfoProviderFactory
from clarin.sru.server.config import SRUServerConfig
from clarin.sru.server.config import SRUServerConfigKey
from clarin.sru.server.server import SRUSearchEngine
from clarin.sru.server.server import SRUServer

if typing.TYPE_CHECKING:
    # import typing_extensions as te
    # wsgiref.types  # only 3.11
    # from _typeshed.wsgi import WSGIApplication
    from _typeshed.wsgi import StartResponse
    from _typeshed.wsgi import WSGIEnvironment


LOGGER = logging.getLogger(__name__)


# ---------------------------------------------------------------------------


class SRUServerApp:  # WSGIApplication
    def __init__(
        self,
        SRUSearchEngine_clazz: Union[Type[SRUSearchEngine], SRUSearchEngine],
        config_file: str,
        params: Dict[Union[SRUServerConfigKey, str], str],
        develop: bool = False,
    ):
        self.SRUSearchEngine_clazz = SRUSearchEngine_clazz
        self.config_file = config_file
        self.params = dict(params)
        self.develop = develop

        self.init()

    # ----------------------------------------------------

    def set_default_params(self) -> None:
        # set defaults
        if self.develop:
            LOGGER.warning("Using >develop< mode!")

            def _set_default(name, value):
                if name not in self.params:
                    self.params[name] = value
                    LOGGER.warning(
                        "Using default '%s' for parameter '%s'"
                        ", because it was not defined in configuration",
                        value,
                        name,
                    )
                    LOGGER.warning("THIS IS NOT RECOMMENDED FOR PRODUCTION DEPLOYMENT!")

            _set_default(SRUServerConfigKey.SRU_TRANSPORT, "http")
            _set_default(SRUServerConfigKey.SRU_HOST, "127.0.0.1")
            _set_default(SRUServerConfigKey.SRU_PORT, "8080")
            _set_default(SRUServerConfigKey.SRU_DATABASE, "/")

            # allowed indented responses
            self.params.setdefault(
                SRUServerConfigKey.SRU_ALLOW_OVERRIDE_INDENT_RESPONSE, "true"
            )

        # set to SRU 2.0 version
        self.params.setdefault(
            SRUServerConfigKey.SRU_SUPPORTED_VERSION_DEFAULT, SRUVersion.VERSION_2_0
        )
        self.params.setdefault(
            SRUServerConfigKey.SRU_SUPPORTED_VERSION_MAX, SRUVersion.VERSION_2_0
        )

    def init(self) -> None:
        # Set some defaults (aka "plug and play" for development deployment)
        # Override those for a production deployment through the params parameter
        self.set_default_params()

        # now go ahead and setup everything ...

        try:
            # parse SRU server configuration
            config = SRUServerConfig.parse(self.params, self.config_file)

            # create an instance of the search engine ...
            LOGGER.debug(
                "Creating new search engine from class %s", self.SRUSearchEngine_clazz
            )
            # TODO: create from package.class name (str) ?
            try:
                if not self.SRUSearchEngine_clazz:
                    raise TypeError("SRUSearchEngine_clazz is None")

                # check that of class type and correct class
                if isinstance(self.SRUSearchEngine_clazz, type):
                    if SRUSearchEngine not in self.SRUSearchEngine_clazz.mro():
                        raise ValueError(
                            "Parameter 'SRUSearchEngine_clazz' not of type SRUSearchEngine"
                        )

                    self.search_engine = self.SRUSearchEngine_clazz()
                else:
                    if not isinstance(self.SRUSearchEngine_clazz, SRUSearchEngine):
                        raise ValueError(
                            "Invalid value for 'SRUSearchEngine_clazz' argument."
                        )

                    # got instance of class instead of type
                    self.search_engine = self.SRUSearchEngine_clazz
            except Exception as ex:
                raise SRUConfigException("Error creating search engine") from ex

            # initialize search engine ...
            qpr_builder = SRUQueryParserRegistry.Builder()
            self.search_engine.init(config, qpr_builder, self.params)

            # create authentication provider
            auth_provider: Optional[SRUAuthenticationInfoProvider] = None
            if isinstance(self.search_engine, SRUAuthenticationInfoProviderFactory):
                LOGGER.debug("Creating new authentication info provider")
                auth_provider = self.search_engine.create_SRUAuthenticationInfoProvider(
                    self.params
                )

            # finally create the sru server ...
            self.server = SRUServer(
                config=config,
                query_parsers=qpr_builder.build(),
                search_engine=self.search_engine,
                authentication_info_provider=auth_provider,
            )

        except SRUConfigException as ex:
            raise RuntimeError(
                f"error configuring or inializing the server: {str(ex)}"
            ) from ex

    def destroy(self) -> None:
        """Destroy the SRU server application"""
        if self.search_engine:
            self.search_engine.destroy()

    # ----------------------------------------------------

    def wsgi_app(
        self, environ: "WSGIEnvironment", start_response: "StartResponse"
    ) -> Iterable[bytes]:
        request = Request(environ)
        response = Response()
        self.server.handle_request(request, response)
        return response(environ, start_response)

    def __call__(
        self, environ: "WSGIEnvironment", start_response: "StartResponse"
    ) -> Iterable[bytes]:
        return self.wsgi_app(environ, start_response)


# ---------------------------------------------------------------------------



import pytest

from clarin.sru.constants import SRUDiagnostics
from clarin.sru.exception import SRUException

# ---------------------------------------------------------------------------


def test_exception_init():
    with pytest.raises(TypeError) as exc_info:
        raise SRUException()

    with pytest.raises(ValueError) as exc_info:
        raise SRUException(" ")

    with pytest.raises(ValueError) as exc_info:
        raise SRUException("")


def test_exception_chaining():
    uri = SRUDiagnostics.GENERAL_SYSTEM_ERROR.value

    with pytest.raises(
        SRUException, match=SRUDiagnostics.GENERAL_SYSTEM_ERROR.description
    ) as exc_info:
        raise SRUException(uri)

    with pytest.raises(
        SRUException, match=SRUDiagnostics.GENERAL_SYSTEM_ERROR.description
    ) as exc_info:
        try:
            raise Exception("test124")
        except Exception as ex:
            raise SRUException(uri)

    with pytest.raises(SRUException, match="test124") as exc_info:
        try:
            raise Exception("test124")
        except Exception as ex:
            raise SRUException(uri) from ex

    with pytest.raises(SRUException, match="abc98") as exc_info:
        try:
            raise Exception("test124")
        except Exception as ex:
            raise SRUException(uri, message="abc98")

    with pytest.raises(SRUException, match="abc98") as exc_info:
        try:
            raise Exception("test124")
        except Exception as ex:
            raise SRUException(uri, message="abc98") from ex

    # invalid uri
    with pytest.raises(SRUException, match=None) as exc_info:
        raise SRUException("abc")

    with pytest.raises(SRUException, match="ttt2") as exc_info:
        raise SRUException("abc", message="ttt2")

    # missing
    with pytest.raises(TypeError):
        raise SRUException(None)
    with pytest.raises(ValueError):
        raise SRUException("")
    with pytest.raises(ValueError):
        raise SRUException("   \n ")


# ---------------------------------------------------------------------------



# try:
#     import pkg_resources
#
#     pkg_resources.declare_namespace(__name__)
# except ImportError:
#     import pkgutil
#
#     __path__ = pkgutil.extend_path(__path__, __name__)

# https://pawamoy.github.io/posts/plugins-as-python-native-namespace-packages/
# https://github.com/microsoft/pyright/issues/2882

__path__ = __import__("pkgutil").extend_path(__path__, __name__)



import importlib.resources
import io
import os
from dataclasses import dataclass
from dataclasses import field
from enum import Enum
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

from lxml import etree

from ..constants import SRURecordPacking
from ..constants import SRURecordXmlEscaping
from ..constants import SRUVersion
from ..exception import SRUConfigException

# ---------------------------------------------------------------------------


class LegacyNamespaceMode(str, Enum):
    LOC = "loc"
    OASIS = "oasis"


@dataclass(frozen=True)
class LocalizedString:
    value: str
    lang: str
    primary: bool = False


# ---------------------------------------------------------------------------


DEFAULT_SRU_VERSION_MIN = SRUVersion.VERSION_1_1
DEFAULT_SRU_VERSION_MAX = SRUVersion.VERSION_1_2
DEFAULT_LEGACY_NAMESPACE_MODE = LegacyNamespaceMode.LOC

DEFAULT_NUMBER_OF_RECORDS = 100
DEFAULT_MAXIMUM_RECORDS = 250
DEFAULT_NUMBER_OF_TERMS = 250
DEFAULT_MAXIMUM_TERMS = 500
DEFAULT_RESPONSE_BUFFER_SIZE = 64 * 1024

CONFIG_FILE_NAMESPACE_PREFIX = "sru"
CONFIG_FILE_NAMESPACE_URI = "http://www.clarin.eu/sru-server/1.0/"
CONFIG_FILE_SCHEMA_PACKAGE = "clarin.sru.xml"
CONFIG_FILE_SCHEMA_URL = "sru-server-config.xsd"

# from: https://docs.oracle.com/javase/7/docs/api/javax/xml/XMLConstants.html
XML_NS_PREFIX = "xml"
XML_NS_URI = "http://www.w3.org/XML/1998/namespace"
NULL_NS_URI = ""
W3C_XML_SCHEMA_NS_URI = "http://www.w3.org/2001/XMLSchema"


# ---------------------------------------------------------------------------


class SRUServerConfigKey(str, Enum):
    SRU_SUPPORTED_VERSION_MIN = "eu.clarin.sru.server.sruSupportedVersionMin"
    """Parameter constant for setting the minimum supported SRU
    version for this SRU server. Must be smaller or equal to
    `SRU_SUPPORTED_VERSION_MAX`.

    Valid values: "``1.1``", "``1.2``" or " ``2.0``" (without
    quotation marks)
    """

    SRU_SUPPORTED_VERSION_MAX = "eu.clarin.sru.server.sruSupportedVersionMax"
    """Parameter constant for setting the maximum supported SRU
    version for this SRU server. Must be larger or equal to
    `SRU_SUPPORTED_VERSION_MIN`.

    Valid values: "``1.1``", "``1.2``" or "``2.0``" (without
    quotation marks)
    """

    SRU_SUPPORTED_VERSION_DEFAULT = "eu.clarin.sru.server.sruSupportedVersionDefault"
    """Parameter constant for setting the default SRU version for
    this SRU server, e.g. for an **Explain** request without explicit
    version.

    Must not me less than `SRU_SUPPORTED_VERSION_MIN` or larger than
    `SRU_SUPPORTED_VERSION_MAX`. Defaults to `SRU_SUPPORTED_VERSION_MAX`.

    Valid values: "``1.1``", "``1.2``" or "``2.0``" (without
    quotation marks)
    """

    SRU_LEGACY_NAMESPACE_MODE = "eu.clarin.sru.server.legacyNamespaceMode"
    """Parameter constant for setting the namespace URIs for SRU 1.1
    and SRU 1.2.

    Valid values: "``loc``" for Library Of Congress URI or "``oasis``"
    for OASIS URIs (without quotation marks).
    """

    SRU_TRANSPORT = "eu.clarin.sru.server.transport"
    """Parameter constant for configuring the transports for this SRU
    server.

    Valid values: "``http``", "``https``" or "``http https``"
    (without quotation marks)

    Used as part of the **Explain** response.
    """

    SRU_HOST = "eu.clarin.sru.server.host"
    """Parameter constant for configuring the host of this SRU server.

    Valid values: any fully qualified hostname,
    e.g. ``sru.example.org``.

    Used as part of the **Explain** response.
    """

    SRU_PORT = "eu.clarin.sru.server.port"
    """Parameter constant for configuring the port number of this SRU
    server.

    Valid values: number between 1 and 65535 (typically 80 or 8080)

    Used as part of the **Explain** response.
    """

    SRU_DATABASE = "eu.clarin.sru.server.database"
    """Parameter constant for configuring the database of this SRU
    server. This is usually the path component of the SRU servers URI.

    Valid values: typically the path component if the SRU server URI.

    Used as part of the **Explain** response.
    """

    SRU_NUMBER_OF_RECORDS = "eu.clarin.sru.server.numberOfRecords"
    """Parameter constant for configuring the **default** number of
    records the SRU server will provide in the response to a
    **searchRetrieve** request if the client does not provide this
    value.

    Valid values: a integer greater than 0 (default value is 100)
    """

    SRU_MAXIMUM_RECORDS = "eu.clarin.sru.server.maximumRecords"
    """Parameter constant for configuring the **maximum** number of
    records the SRU server will support in the response to a
    **searchRetrieve** request. If a client requests more records,
    the number will be limited to this value.

    Valid values: a integer greater than 0 (default value is 250)
    """

    SRU_NUMBER_OF_TERMS = "eu.clarin.sru.server.numberOfTerms"
    """Parameter constant for configuring the **default** number of
    terms the SRU server will provide in the response to a **scan**
    request if the client does not provide this value.

    Valid values: a integer greater than 0 (default value is 250)
    """

    SRU_MAXIMUM_TERMS = "eu.clarin.sru.server.maximumTerms"
    """Parameter constant for configuring the **maximum** number of
    terms the SRU server will support in the response to a **scan**
    request. If a client requests more records, the number will be
    limited to this value.

    Valid values: a integer greater than 0 (default value is 500)
    """

    SRU_ECHO_REQUESTS = "eu.clarin.sru.server.echoRequests"
    """Parameter constant for configuring, if the SRU server will
    echo the request.

    Valid values: ``true`` or ``false``
    """

    SRU_INDENT_RESPONSE = "eu.clarin.sru.server.indentResponse"
    """Parameter constant for configuring, if the SRU server
    pretty-print the XML response. Setting this parameter can be
    useful for manual debugging of the XML response, however it is
    **not recommended** for production setups.

    Valid values: any integer greater or equal to ``-1`` (default)
    and less or equal to ``8``
    """

    SRU_ALLOW_OVERRIDE_MAXIMUM_RECORDS = (
        "eu.clarin.sru.server.allowOverrideMaximumRecords"
    )
    """Parameter constant for configuring, if the SRU server will
    allow the client to override the maximum number of records the
    server supports. This parameter is solely intended for debugging
    and setting it to ``true`` is **strongly** discouraged for
    production setups.

    Valid values: ``true`` or ``false`` (default)
    """

    SRU_ALLOW_OVERRIDE_MAXIMUM_TERMS = "eu.clarin.sru.server.allowOverrideMaximumTerms"
    """Parameter constant for configuring, if the SRU server will
    allow the client to override the maximum number of terms the
    server supports. This parameter is solely intended for debugging
    and setting it to ``true`` it is **strongly** discouraged for
    production setups.

    Valid values: ``true`` or ``false`` (default)
    """

    SRU_ALLOW_OVERRIDE_INDENT_RESPONSE = (
        "eu.clarin.sru.server.allowOverrideIndentResponse"
    )
    """Parameter constant for configuring, if the SRU server will
    allow the client to override the pretty-printing setting of the
    server. This parameter is solely intended for debugging and
    setting it to ``true`` it is **strongly** discouraged for
    production setups.

    Valid values: ``true`` or ``false`` (default)
    """

    # TODO: docstring; needed?
    SRU_RESPONSE_BUFFER_SIZE = "eu.clarin.sru.server.responseBufferSize"
    """Parameter constant for configuring the size of response buffer.
    The Servlet will buffer up to this amount of data before sending
    a response to the client. This value specifies the size of the
    buffer in bytes.

    Valid values: any positive integer (default 65536)
    """


# ---------------------------------------------------------------------------


@dataclass(frozen=True)
class DatabaseInfo:
    title: Optional[List[LocalizedString]] = None
    description: Optional[List[LocalizedString]] = None
    author: Optional[List[LocalizedString]] = None
    extent: Optional[List[LocalizedString]] = None
    history: Optional[List[LocalizedString]] = None
    langUsage: Optional[List[LocalizedString]] = None
    restrictions: Optional[List[LocalizedString]] = None
    subjects: Optional[List[LocalizedString]] = None
    links: Optional[List[LocalizedString]] = None
    implementation: Optional[List[LocalizedString]] = None


@dataclass(frozen=True)
class SchemaInfo:
    identifier: str
    name: str
    location: str
    sort: bool
    retrieve: bool
    title: Optional[List[LocalizedString]] = None


@dataclass(frozen=True)
class IndexInfo:
    @dataclass(frozen=True)
    class Set:
        identifier: str
        name: str
        title: Optional[List[LocalizedString]] = None

    @dataclass(frozen=True)
    class Index:
        @dataclass(frozen=True)
        class Map:
            primary: bool
            set: str
            name: str

        can_search: bool
        can_scan: bool
        can_sort: bool
        maps: Optional[List[Map]] = None
        title: Optional[List[LocalizedString]] = None

    sets: Optional[List[Set]] = None
    indexes: Optional[List[Index]] = None


# ---------------------------------------------------------------------------


@dataclass
class SRUServerConfig:
    """SRU server configuration.

    The XML configuration file must validate against the
    ``sru-server-config.xsd`` W3C schema bundled with the package and
    need to have the ``http://www.clarin.eu/sru-server/1.0/`` XML
    namespace.
    """

    min_version: SRUVersion
    max_version: SRUVersion
    default_version: SRUVersion
    legacy_namespace_mode: LegacyNamespaceMode
    transport: str
    host: str
    port: int
    database: str
    number_of_records: int
    maximum_records: int
    number_of_terms: int
    maximum_terms: int
    echo_requests: bool
    indent_response: int
    response_buffer_size: int
    allow_override_maximum_records: bool
    allow_override_maximum_terms: bool
    allow_override_indent_response: bool
    base_url: str = field(init=False)
    database_info: DatabaseInfo
    index_info: IndexInfo
    schema_info: Optional[List[SchemaInfo]] = None

    def __post_init__(self):
        self.base_url = f"{self.host}{':' + str(self.port) if self.port != 80 else ''}/{self.database}"
        if not self.schema_info:
            self.schema_info = None

    @property
    def default_record_xml_escaping(self) -> SRURecordXmlEscaping:
        return SRURecordXmlEscaping.XML

    @property
    def default_record_packing(self) -> SRURecordPacking:
        return SRURecordPacking.PACKED

    def get_record_schema_identifier(self, record_schema_name: str) -> Optional[str]:
        if record_schema_name is not None:
            if self.schema_info:
                for schema in self.schema_info:
                    if schema.name == record_schema_name:
                        return schema.identifier
        return None

    def get_record_schema_name(self, schema_identifier: str) -> Optional[str]:
        if schema_identifier is not None:
            if self.schema_info:
                for schema in self.schema_info:
                    if schema.identifier == schema_identifier:
                        return schema.name
        return None

    def find_schema_info(self, value: str) -> Optional[SchemaInfo]:
        if value is not None:
            if self.schema_info:
                for schema in self.schema_info:
                    if schema.identifier == value or schema.name == value:
                        return schema
        return None

    @staticmethod
    def find_set_by_name(
        sets: List[IndexInfo.Set], name: str
    ) -> Optional[IndexInfo.Set]:
        for set in sets:
            if set.name == name:
                return set
        return None

    @staticmethod
    def fromparams(
        params: Dict[str, str],
        database_info: DatabaseInfo,
        index_info: Optional[IndexInfo] = None,
        schema_info: Optional[List[SchemaInfo]] = None,
    ) -> "SRUServerConfig":
        """Creates an SRU configuration object with default values
        and overrides from **params**.

        Args:
            params: additional settings
            database_info: optinal `DatabaseInfo`
            index_info: optinal `IndexInfo`
            schema_info: optional list `SchemaInfo`

        Returns:
            SRUServerConfig: a initialized `SRUEndpointConfig` instance

        Raises:
            `TypeError`: if **params** is None
            `SRUConfigException`: if an error occurred
        """
        if params is None:
            raise TypeError("params is None")

        # NOTE: maybe some more validation? but everything could possibly be empty
        if database_info is None:
            database_info = DatabaseInfo()
        if index_info is None:
            index_info = IndexInfo()

        try:
            # fetch parameters more parameters (usually passed from
            # environment)

            min_version = SRUServerConfig.parse_version(
                params,
                SRUServerConfigKey.SRU_SUPPORTED_VERSION_MIN,
                False,
                DEFAULT_SRU_VERSION_MIN,
            )
            max_version = SRUServerConfig.parse_version(
                params,
                SRUServerConfigKey.SRU_SUPPORTED_VERSION_MAX,
                False,
                DEFAULT_SRU_VERSION_MAX,
            )
            if max_version.version_number < min_version.version_number:
                raise SRUConfigException(
                    f"parameter value '{SRUServerConfigKey.SRU_SUPPORTED_VERSION_MAX}'"
                    f" ({max_version.version_string}) must be equal or larger than value"
                    f"of parameter '{SRUServerConfigKey.SRU_SUPPORTED_VERSION_MIN}'"
                    f" ({min_version.version_string})"
                )

            default_version = SRUServerConfig.parse_version(
                params,
                SRUServerConfigKey.SRU_SUPPORTED_VERSION_DEFAULT,
                False,
                max_version,
            )
            if (
                default_version.version_number < min_version.version_number
                or default_version.version_number > max_version.version_number
            ):
                raise SRUConfigException(
                    f"parameter value '{SRUServerConfigKey.SRU_SUPPORTED_VERSION_DEFAULT}'"
                    f" ({default_version.version_string}) must be between value of parameter"
                    f" '{SRUServerConfigKey.SRU_SUPPORTED_VERSION_MIN}'"
                    f" ({min_version.version_string}) and"
                    f" '{SRUServerConfigKey.SRU_SUPPORTED_VERSION_MAX}'"
                    f" ({max_version.version_string})"
                )

            legacy_namespace_mode = DEFAULT_LEGACY_NAMESPACE_MODE
            mode = params.get(SRUServerConfigKey.SRU_LEGACY_NAMESPACE_MODE)
            if mode is not None and isinstance(mode, str) and not mode.strip():
                if LegacyNamespaceMode.LOC.value == mode:
                    legacy_namespace_mode = LegacyNamespaceMode.LOC
                elif LegacyNamespaceMode.OASIS.value == mode:
                    legacy_namespace_mode = LegacyNamespaceMode.OASIS
                else:
                    raise SRUConfigException(
                        f"invalid value for parameter {SRUServerConfigKey.SRU_LEGACY_NAMESPACE_MODE}: {mode}"
                    )

            transport = params.get(SRUServerConfigKey.SRU_TRANSPORT)
            if (
                transport is None
                or not isinstance(transport, str)
                or not transport.strip()
            ):
                raise SRUConfigException(
                    f"parameter {SRUServerConfigKey.SRU_TRANSPORT} is mandatory"
                )
            transport_parts = transport.strip().lower().split()
            for tr in transport_parts:
                if tr not in ("http", "https"):
                    raise SRUConfigException(f"unsupported transport {tr}")
            transport = " ".join(transport_parts)

            host = params.get(SRUServerConfigKey.SRU_HOST)
            if host is None or not isinstance(host, str) or not host.strip():
                raise SRUConfigException(
                    f"parameter {SRUServerConfigKey.SRU_HOST} is mandatory"
                )

            port = SRUServerConfig.parse_int(
                params, SRUServerConfigKey.SRU_PORT, True, -1, 1, 65535
            )

            database = params.get(SRUServerConfigKey.SRU_DATABASE)
            if (
                database is None
                or not isinstance(database, str)
                or not database.strip()
            ):
                raise SRUConfigException(
                    f"parameter {SRUServerConfigKey.SRU_DATABASE} is mandatory"
                )
            database = database.lstrip("/")

            number_of_records = SRUServerConfig.parse_int(
                params,
                SRUServerConfigKey.SRU_NUMBER_OF_RECORDS,
                False,
                DEFAULT_NUMBER_OF_RECORDS,
                1,
                -1,
            )

            maximum_records = SRUServerConfig.parse_int(
                params,
                SRUServerConfigKey.SRU_MAXIMUM_RECORDS,
                False,
                DEFAULT_MAXIMUM_RECORDS,
                number_of_records,
                -1,
            )

            number_of_terms = SRUServerConfig.parse_int(
                params,
                SRUServerConfigKey.SRU_NUMBER_OF_TERMS,
                False,
                DEFAULT_NUMBER_OF_TERMS,
                0,
                -1,
            )

            maximum_terms = SRUServerConfig.parse_int(
                params,
                SRUServerConfigKey.SRU_MAXIMUM_TERMS,
                False,
                DEFAULT_MAXIMUM_TERMS,
                number_of_terms,
                -1,
            )

            echo_requests = SRUServerConfig.parse_bool(
                params, SRUServerConfigKey.SRU_ECHO_REQUESTS, False, True
            )

            indent_response = SRUServerConfig.parse_int(
                params, SRUServerConfigKey.SRU_INDENT_RESPONSE, False, -1, -1, 8
            )

            allow_override_maximum_records = SRUServerConfig.parse_bool(
                params,
                SRUServerConfigKey.SRU_ALLOW_OVERRIDE_MAXIMUM_RECORDS,
                False,
                False,
            )

            allow_override_maximum_terms = SRUServerConfig.parse_bool(
                params,
                SRUServerConfigKey.SRU_ALLOW_OVERRIDE_MAXIMUM_TERMS,
                False,
                False,
            )

            allow_override_indent_response = SRUServerConfig.parse_bool(
                params,
                SRUServerConfigKey.SRU_ALLOW_OVERRIDE_INDENT_RESPONSE,
                False,
                False,
            )

            response_buffer_size = SRUServerConfig.parse_int(
                params,
                SRUServerConfigKey.SRU_RESPONSE_BUFFER_SIZE,
                False,
                DEFAULT_RESPONSE_BUFFER_SIZE,
                0,
                -1,
            )

            return SRUServerConfig(
                min_version=min_version,
                max_version=max_version,
                default_version=default_version,
                legacy_namespace_mode=legacy_namespace_mode,
                transport=transport,
                host=host,
                port=port,
                database=database,
                number_of_records=number_of_records,
                maximum_records=maximum_records,
                number_of_terms=number_of_terms,
                maximum_terms=maximum_terms,
                echo_requests=echo_requests,
                indent_response=indent_response,
                response_buffer_size=response_buffer_size,
                allow_override_maximum_records=allow_override_maximum_records,
                allow_override_maximum_terms=allow_override_maximum_terms,
                allow_override_indent_response=allow_override_indent_response,
                database_info=database_info,
                index_info=index_info,
                schema_info=schema_info,
            )
        except Exception as ex:
            raise SRUConfigException("error building configuration object") from ex

    @staticmethod
    def parse(
        params: Dict[str, str], config_file: Union[io.BytesIO, os.PathLike, str]
    ) -> "SRUServerConfig":
        """Parse a SRU server XML configuration file and create an
        configuration object from it.

        Args:
            params: additional settings
            config_file: an ``URL`` pointing to the XML configuration
                file

        Returns:
            SRUServerConfig: a initialized `SRUEndpointConfig` instance

        Raises:
            `TypeError`: if **params** or **configFile** is None
            `SRUConfigException`: if an error occurred
        """
        if params is None:
            raise TypeError("params is None")
        if config_file is None:
            raise TypeError("config_file is None")

        try:
            doc = SRUServerConfig.load_config_file(config_file)

            database_info = SRUServerConfig._build_DatabaseInfo(doc)
            index_info = SRUServerConfig._build_IndexInfo(doc)
            schema_info = SRUServerConfig._build_SchemaInfo(doc)
        except Exception as ex:
            raise SRUConfigException("error reading configuration file") from ex

        return SRUServerConfig.fromparams(
            params,
            database_info=database_info,
            index_info=index_info,
            schema_info=schema_info,
        )

    @staticmethod
    def load_config_file(
        config_file: Union[io.BytesIO, os.PathLike, str]
    ) -> etree._ElementTree:
        # load validation schema
        if not importlib.resources.is_resource(
            CONFIG_FILE_SCHEMA_PACKAGE, CONFIG_FILE_SCHEMA_URL
        ):
            raise SRUConfigException(
                f"cannot open {CONFIG_FILE_SCHEMA_URL} in {CONFIG_FILE_SCHEMA_PACKAGE}"
            )

        with importlib.resources.open_text(
            CONFIG_FILE_SCHEMA_PACKAGE,
            CONFIG_FILE_SCHEMA_URL,
            encoding="utf-8",
            errors="strict",
        ) as fp:
            config_schema_doc = etree.parse(fp)

        config_schema = etree.XMLSchema(config_schema_doc)

        parser = etree.XMLParser(ns_clean=False, remove_comments=True)

        # load config xml
        config_doc: etree._ElementTree = etree.parse(config_file, parser)
        docinfo: etree.DocInfo = config_doc.docinfo
        docinfo.public_id = CONFIG_FILE_NAMESPACE_URI
        docinfo.system_url = CONFIG_FILE_NAMESPACE_URI

        # validate
        config_schema.assertValid(config_doc)

        return config_doc

    @staticmethod
    def parse_version(
        params: Dict[str, str], name: str, mandatory: bool, default: SRUVersion
    ) -> SRUVersion:
        value = params.get(name)
        if value is None or not isinstance(value, str) or not value.strip():
            if mandatory:
                raise SRUConfigException(f"parameter '{name}' is mandatory")
            else:
                return default

        if "1.1" == value:
            return SRUVersion.VERSION_1_1
        if "1.2" == value:
            return SRUVersion.VERSION_1_2
        if "2.0" == value:
            return SRUVersion.VERSION_2_0

        raise SRUConfigException(f"invalid value for parameter '{name}': value")

    @staticmethod
    def parse_int(
        params: Dict[str, str],
        name: str,
        mandatory: bool,
        default: int,
        min: int,
        max: int,
    ) -> int:
        value = params.get(name)
        if value is None or not isinstance(value, str) or not value.strip().isdigit():
            if mandatory:
                raise SRUConfigException(f"parameter '{name}' is mandatory")
            else:
                return default
        try:
            num = int(value)

            # sanity checks
            if min != -1 and max != -1:
                if num < min or num > max:
                    raise SRUConfigException(
                        f"parameter '{name}' must be between {min} and {max}: {num}"
                    )
            else:
                if min != -1 and num < min:
                    raise SRUConfigException(
                        f"parameter '{name}' must be larger than {min}: {num}"
                    )
                if max != -1 and num > max:
                    raise SRUConfigException(
                        f"parameter '{name}' must be smaller than {max}: {num}"
                    )

            return num
        except ValueError as ex:
            import sys

            raise SRUConfigException(
                f"parameter '{name}' must be nummerical and less than {sys.maxsize}: {value}"
            ) from ex

    @staticmethod
    def parse_bool(
        params: Dict[str, str], name: str, mandatory: bool, default: bool
    ) -> bool:
        value = params.get(name)
        if (
            value is None
            or not isinstance(value, str)
            or value.strip().lower() not in ("true", "false")
        ):
            if mandatory:
                raise SRUConfigException(f"parameter '{name}' is mandatory")
            else:
                return default
        return value.strip().lower() == "true"

    @staticmethod
    def _build_DatabaseInfo(doc: etree._ElementTree) -> DatabaseInfo:
        title = SRUServerConfig._build_list(doc, "//sru:databaseInfo/sru:title")
        description = SRUServerConfig._build_list(
            doc, "//sru:databaseInfo/sru:description"
        )
        author = SRUServerConfig._build_list(doc, "//sru:databaseInfo/sru:author")
        extent = SRUServerConfig._build_list(doc, "//sru:databaseInfo/sru:extent")
        history = SRUServerConfig._build_list(doc, "//sru:databaseInfo/sru:history")
        langUsage = SRUServerConfig._build_list(doc, "//sru:databaseInfo/sru:langUsage")
        restrictions = SRUServerConfig._build_list(
            doc, "//sru:databaseInfo/sru:restrictions"
        )
        subjects = SRUServerConfig._build_list(doc, "//sru:databaseInfo/sru:subjects")
        links = SRUServerConfig._build_list(doc, "//sru:databaseInfo/sru:links")
        implementation = SRUServerConfig._build_list(
            doc, "//sru:databaseInfo/sru:implementation"
        )

        return DatabaseInfo(
            title=title,
            description=description,
            author=author,
            extent=extent,
            history=history,
            langUsage=langUsage,
            restrictions=restrictions,
            subjects=subjects,
            links=links,
            implementation=implementation,
        )

    @staticmethod
    def _build_IndexInfo(doc: etree._ElementTree) -> IndexInfo:
        sets: Optional[List[IndexInfo.Set]] = None
        indexes: Optional[List[IndexInfo.Index]] = None

        nodes = doc.xpath(
            "//sru:indexInfo/sru:set",
            namespaces={CONFIG_FILE_NAMESPACE_PREFIX: CONFIG_FILE_NAMESPACE_URI},
        )
        if nodes:
            sets = list()
            for node in nodes:
                identifier = node.get("identifier")
                name = node.get("name")
                if not identifier or not identifier.strip():
                    raise SRUConfigException(
                        "attribute 'identifier' may on element '/indexInfo/set' may not be empty"
                    )
                if not name or not name.strip():
                    raise SRUConfigException(
                        "attribute 'name' may on element '/indexInfo/set' may not be empty"
                    )
                title = SRUServerConfig._from_node_list(
                    node.findall("title", namespaces=node.nsmap)
                )

                sets.append(
                    IndexInfo.Set(identifier=identifier, name=name, title=title)
                )

        nodes = doc.xpath(
            "//sru:indexInfo/sru:index",
            namespaces={CONFIG_FILE_NAMESPACE_PREFIX: CONFIG_FILE_NAMESPACE_URI},
        )
        if nodes:
            indexes = list()
            for node in nodes:
                title = SRUServerConfig._from_node_list(
                    node.findall("title", namespaces=node.nsmap)
                )
                can_search = SRUServerConfig._get_bool_attrib(node, "search", False)
                can_scan = SRUServerConfig._get_bool_attrib(node, "scan", False)
                can_sort = SRUServerConfig._get_bool_attrib(node, "sort", False)

                maps: Optional[List[IndexInfo.Index.Map]] = None
                nodes_map = node.findall("map", namespaces=node.nsmap)
                if nodes_map:
                    maps = list()
                    found_primary = False
                    for node_map in nodes_map:
                        primary = SRUServerConfig._get_bool_attrib(
                            node_map, "primary", False
                        )
                        if primary:
                            if found_primary:
                                raise SRUConfigException(
                                    "only one map may be 'primary' in index"
                                )
                            found_primary = True

                        node_name = node_map.find("name", namespaces=node_map.nsmap)
                        if node_name is not None:
                            map_set = node_name.get("set")
                            map_name = node_name.text
                            if not map_set or not map_set.strip():
                                raise SRUConfigException(
                                    "attribute 'set' on element '/indexInfo/index/map/name' may not be empty"
                                )
                            if map_name is None or not map_name.strip():
                                raise SRUConfigException(
                                    "element '/indexInfo/index/map/name' may not be empty"
                                )

                            # clarin/java code would allow setting None to both set/name
                            # (if there was no schema validation)
                            # so we can assume that when we find a map entry, a name entry
                            # exists with the correct attribute
                            # so we skip adding the map on identation lower --> mypy checks ...
                            maps.append(
                                IndexInfo.Index.Map(
                                    primary=primary, set=map_set, name=map_name
                                )
                            )

                indexes.append(
                    IndexInfo.Index(
                        title=title,
                        can_search=can_search,
                        can_scan=can_scan,
                        can_sort=can_sort,
                        maps=maps,
                    )
                )

            # sanity check (/index/map/name/@set exists in any set/@name)
            if sets:
                for index in indexes:
                    if not index.maps:
                        continue
                    for map in index.maps:
                        if not SRUServerConfig.find_set_by_name(sets, map.set):
                            raise SRUConfigException(
                                f"/index/map/name refers to nonexitsing set ({map.set})"
                            )

        return IndexInfo(sets, indexes)

    @staticmethod
    def _build_SchemaInfo(doc: etree._ElementTree) -> Optional[List[SchemaInfo]]:
        nodes = doc.xpath(
            "//sru:schemaInfo/sru:schema",
            namespaces={CONFIG_FILE_NAMESPACE_PREFIX: CONFIG_FILE_NAMESPACE_URI},
        )
        if not nodes:
            return None

        schemaInfos: List[SchemaInfo] = list()
        for node in nodes:
            identifier = node.get("identifier")
            name = node.get("name")
            location = node.get("location")
            if location is not None and not location.strip():
                location = None
            sort = SRUServerConfig._get_bool_attrib(node, "sort", False)
            retrieve = SRUServerConfig._get_bool_attrib(node, "retrieve", False)
            title = SRUServerConfig._from_node_list(
                node.findall("title", namespaces=node.nsmap)
            )

            schemaInfos.append(
                SchemaInfo(
                    identifier=identifier,
                    name=name,
                    location=location,
                    sort=sort,
                    retrieve=retrieve,
                    title=title,
                )
            )

        return schemaInfos

    @staticmethod
    def _build_list(
        doc: etree._ElementTree, xpath: str
    ) -> Optional[List[LocalizedString]]:
        return SRUServerConfig._from_node_list(
            doc.xpath(
                xpath,
                namespaces={CONFIG_FILE_NAMESPACE_PREFIX: CONFIG_FILE_NAMESPACE_URI},
            )
        )

    @staticmethod
    def _from_node_list(nodes: List[etree._Element]) -> Optional[List[LocalizedString]]:
        if nodes is None or not len(nodes):
            return None

        strings = list()
        found_primary = False
        for node in nodes:
            primary = SRUServerConfig._get_bool_attrib(node, "primary", False)
            if primary:
                if found_primary:
                    raise SRUConfigException(
                        "list may only contain one element as primary"
                    )
                found_primary = True

            strings.append(
                LocalizedString(
                    value=node.text,
                    lang=node.get(etree.QName(XML_NS_URI, "lang")),
                    primary=primary,
                )
            )

        return strings

    @staticmethod
    def _get_bool_attrib(node: etree._Element, local_name: str, default: bool) -> bool:
        value = node.get(local_name)
        if (
            value is None
            or not value.strip()
            or value.strip().lower() not in ("true", "false")
        ):
            return default
        return value.strip().lower() == "true"


# ---------------------------------------------------------------------------



# from: https://pythonbasics.org/webserver/
# from http.server import BaseHTTPRequestHandler, HTTPServer

import io
import logging
from abc import ABCMeta
from abc import abstractmethod
from dataclasses import dataclass
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from xml.sax import SAXException

import cql
from werkzeug.wrappers import Request
from werkzeug.wrappers import Response

from ..constants import RESPONSE_CONTENT_TYPE
from ..constants import RESPONSE_ENCODING
from ..constants import SRU_DIAGNOSTIC_RECORD_SCHEMA
from ..constants import SRUDiagnostics
from ..constants import SRUOperation
from ..constants import SRUQueryType
from ..constants import SRURecordPacking
from ..constants import SRURecordXmlEscaping
from ..constants import SRUVersion
from ..diagnostic import SRUDiagnostic
from ..diagnostic import SRUDiagnosticList
from ..exception import SRUException
from ..queryparser import SRUQuery
from ..queryparser import SRUQueryParserRegistry
from ..xml.writer import SRUXMLStreamWriter
from .auth import SRUAuthenticationInfoProvider
from .config import DatabaseInfo
from .config import IndexInfo
from .config import LegacyNamespaceMode
from .config import LocalizedString
from .config import SchemaInfo
from .config import SRUServerConfig
from .request import SRURequest
from .request import SRURequestImpl
from .result import SRUExplainResult
from .result import SRUScanResultSet
from .result import SRUSearchResultSet

# ---------------------------------------------------------------------------


LOGGER = logging.getLogger(__name__)


SRU_NS = "http://www.loc.gov/zing/srw/"
SRU_PREFIX = "sru"
SRU_RESPONSE_NS = "http://docs.oasis-open.org/ns/search-ws/sruResponse"
SRU_RESPONSE_PREFIX = "sruResponse"
SRU_SCAN_NS = "http://docs.oasis-open.org/ns/search-ws/scan"
SRU_SCAN_PREFIX = "scan"
SRU_DIAGNOSIC_NS = "http://docs.oasis-open.org/ns/search-ws/diagnostic"
SRU_DIAGNOSTIC_PREFIX = "diag"
SRU_EXPLAIN_NS = "http://explain.z3950.org/dtd/2.0/"
SRU_EXPLAIN_PREFIX = "zr"
SRU_XCQL_NS = "http://docs.oasis-open.org/ns/search-ws/xcql"


@dataclass(frozen=True)
class SRUNamespaces:
    """Interface for decoupling SRU namespaces from implementation to
    allow to support SRU 1.1/1.2 and SRU 2.0."""

    response_NS: str
    """The namespace URI for encoding **explain** and
    **searchRetrieve** operation responses."""

    response_prefix: str
    """The namespace prefix for encoding **explain** and
    **searchRetrieve**"""

    scan_NS: str
    """The namespace URI for encoding **scan** operation responses."""

    scan_prefix: str
    """The namespace prefix for encoding **scan** operation
    responses."""

    diagnostic_NS: str
    """The namespace URI for encoding SRU diagnostics."""

    XCQL_NS: str
    """The namespace URI for encoding XCQL fragments"""

    diagnostic_prefix: str = SRU_DIAGNOSTIC_PREFIX
    """The namespace prefix for encoding SRU diagnostics."""

    explain_NS: str = SRU_EXPLAIN_NS
    """The namespace URI for encoding explain record data fragments."""

    explain_prefix: str = SRU_EXPLAIN_PREFIX
    """The namespace prefix for encoding explain record data
    fragments."""

    # ----------------------------------------------------

    @staticmethod
    def for_legacy_LOC() -> "SRUNamespaces":
        return SRUNamespaces(
            response_NS=SRU_NS,
            response_prefix=SRU_PREFIX,
            scan_NS=SRU_NS,
            scan_prefix=SRU_PREFIX,
            diagnostic_NS="http://www.loc.gov/zing/srw/diagnostic/",
            XCQL_NS="http://www.loc.gov/zing/cql/xcql/",
        )

    @staticmethod
    def for_1_2_OASIS() -> "SRUNamespaces":
        return SRUNamespaces(
            response_NS=SRU_RESPONSE_NS,
            response_prefix=SRU_RESPONSE_PREFIX,
            scan_NS=SRU_SCAN_NS,
            scan_prefix=SRU_SCAN_PREFIX,
            diagnostic_NS=SRU_DIAGNOSIC_NS,
            XCQL_NS=SRU_XCQL_NS,
        )

    @staticmethod
    def for_2_0() -> "SRUNamespaces":
        return SRUNamespaces(
            response_NS=SRU_RESPONSE_NS,
            response_prefix=SRU_RESPONSE_PREFIX,
            scan_NS=SRU_SCAN_NS,
            scan_prefix=SRU_SCAN_PREFIX,
            diagnostic_NS=SRU_DIAGNOSIC_NS,
            XCQL_NS=SRU_XCQL_NS,
        )

    @staticmethod
    def get_namespaces(
        version: SRUVersion, legacy_ns_mode: LegacyNamespaceMode
    ) -> "SRUNamespaces":
        if version is None:
            raise TypeError("version is None")

        if version in (SRUVersion.VERSION_1_1, SRUVersion.VERSION_1_2):
            if legacy_ns_mode == LegacyNamespaceMode.LOC:
                return SRUNamespaces.for_legacy_LOC()
            if legacy_ns_mode == LegacyNamespaceMode.OASIS:
                return SRUNamespaces.for_1_2_OASIS()
            raise ValueError(f"invalid legacy mode: {legacy_ns_mode}")
        if version == SRUVersion.VERSION_2_0:
            return SRUNamespaces.for_2_0()
        raise ValueError(f"invalid version: {version}")


# ---------------------------------------------------------------------------


# TODO: update docstring when SRUServerApp (WSGI) is ready
# SRUSearchEngine + SRUSearchEngineBase
class SRUSearchEngine(metaclass=ABCMeta):
    """Interface for connecting the SRU protocol implementation to an
    actual search engine. Base class required for an `SRUSearchEngine`
    implementation to be used with the `SRUServerApp`.

    Implementing the `explain` and `scan` is optional, but implementing
    `search` is mandatory.

    The implementation of these methods **must** be thread-safe.
    """

    @abstractmethod
    def explain(
        self,
        config: SRUServerConfig,
        request: SRURequest,
        diagnostics: SRUDiagnosticList,
    ) -> Optional[SRUExplainResult]:
        """Handle an **explain** operation. Implementing this method
        is optional, but is required, if the **writeExtraResponseData**
        block of the SRU response needs to be filled. The arguments
        for this operation are provides by the `SRURequest` object.

        The implementation of this method **must** be thread-safe.

        Args:
            config: the `SRUEndpointConfig` object that contains the
                endpoint configuration
            request: the `SRURequest` object that contains the
                request made to the endpoint
            diagnostics: the `SRUDiagnosticList` object for storing
                non-fatal diagnostics

        Returns:
            SRUExplainResult: a `SRUExplainResult` object or ``None``
                if the search engine does not want to provide
                `write_extra_response_data`

        Raises:
            `SRUException`: if an fatal error occurred
        """

    @abstractmethod
    def search(
        self,
        config: SRUServerConfig,
        request: SRURequest,
        diagnostics: SRUDiagnosticList,
    ) -> SRUSearchResultSet:
        """Handle a **searchRetrieve** operation. Implementing this
        method is mandatory. The arguments for this operation are
        provides by the `SRURequest` object.

        The implementation of this method **must** be thread-safe.

        Args:
            config: the `SRUEndpointConfig` object that contains the
                endpoint configuration
            request: the `SRURequest` object that contains the
                request made to the endpoint
            diagnostics: the `SRUDiagnosticList` object for storing
                non-fatal diagnostics

        Returns:
            SRUSearchResultSet: a `SRUSearchResultSet` object

        Raises:
            `SRUException`: if an fatal error occurred
        """

    @abstractmethod
    def scan(
        self,
        config: SRUServerConfig,
        request: SRURequest,
        diagnostics: SRUDiagnosticList,
    ) -> Optional[SRUScanResultSet]:
        """Handle a **scan** operation. Implementing this method is
        optional. If you don't need to handle the **scan** operation,
        just return ``None`` and the SRU server will return the
        appropiate diagnostic to the client. The arguments for this
        operation are provides by the `SRURequest` object.

        The implementation of this method **must** be thread-safe.

        Args:
            config: the `SRUEndpointConfig` object that contains the
                endpoint configuration
            request: the `SRURequest` object that contains the
                request made to the endpoint
            diagnostics: the `SRUDiagnosticList` object for storing
                non-fatal diagnostics

        Returns:
            SRUScanResultSet: a `SRUScanResultSet` object or ``None``
                if this operation is not supported by this search
                engine

        Raises:
            `SRUException`: if an fatal error occurred
        """

    # ----------------------------------------------------

    def init(
        self,
        config: SRUServerConfig,
        query_parser_registry_builder: SRUQueryParserRegistry.Builder,
        params: Dict[str, str],
    ) -> None:
        """Initialize the search engine.

        Args:
            config: the `SRUServerConfig` object for this search engine
            query_parser_registry_builder: the `SRUQueryParserRegistry.Builder`
                object to be used for this search engine. Use to register
                additional query parsers with the `SRUServer`
            params: additional parameters from the server

        Raises:
            SRUConfigException: an error occurred during initialization
                of the search engine
        """

    def destroy(self) -> None:
        """Destroy the search engine. Use this method for any cleanup
        the search engine needs to perform upon termination.
        """


# ---------------------------------------------------------------------------


class SRUServer:
    """SRU/CQL protocol implementation for the server-side (SRU/S).
    This class implements SRU/CQL version 1.1 and and 1.2.

    See also:
        SRU/CQL protocol 1.2: http://www.loc.gov/standards/sru/
    """

    def __init__(
        self,
        config: SRUServerConfig,
        query_parsers: SRUQueryParserRegistry,
        search_engine: SRUSearchEngine,
        authentication_info_provider: Optional[SRUAuthenticationInfoProvider] = None,
    ):
        if config is None:
            raise TypeError("config is None")
        if query_parsers is None:
            raise TypeError("query_parsers is None")
        if search_engine is None:
            raise TypeError("search_engine is None")

        self.config = config
        self.query_parsers = query_parsers
        self.search_engine = search_engine
        self.authentication_info_provider = authentication_info_provider

    # ----------------------------------------------------

    def handle_request(self, request: Request, response: Response):
        """Handle a SRU request."""

        req = SRURequestImpl(
            self.config,
            self.query_parsers,
            request,
            authentication_info_provider=self.authentication_info_provider,
        )
        try:
            # set response properties
            response.content_type = RESPONSE_CONTENT_TYPE
            response.content_encoding = RESPONSE_ENCODING
            response.status = 200  # type: ignore
            # TODO: buffer size? stream
            # self.config.response_buffer_size

            try:
                if req.check_parameters():
                    operation = req.get_operation()
                    if operation == SRUOperation.EXPLAIN:
                        self.explain(req, response)
                    elif operation == SRUOperation.SCAN:
                        self.scan(req, response)
                    elif operation == SRUOperation.SEARCH_RETRIEVE:
                        self.search(req, response)

                else:
                    # (some) parameters are malformed, send error
                    buf = io.StringIO()
                    out = self._create_XML_builder(
                        buf, SRURecordXmlEscaping.XML, False, req.get_indent_response()
                    )
                    ns = SRUNamespaces.get_namespaces(
                        req.get_version(), self.config.legacy_namespace_mode
                    )
                    self._write_fatal_error(out, ns, req, req.diagnostics)
                    response.set_data(buf.getvalue())
            except SAXException as ex:
                LOGGER.error("An error occurred while serializing response", ex)
                raise SRUException(
                    SRUDiagnostics.GENERAL_SYSTEM_ERROR,
                    message="An error occurred while serializing response.",
                ) from ex
            except Exception:
                # Well, can't really do anything useful here ...
                LOGGER.error("An unexpected exception occurred", exc_info=True)
        except SRUException as ex:
            # check if response.isCommitted does not exist for werkzeug
            if LOGGER.isEnabledFor(logging.INFO):
                message = ex.get_diagnostic().message
                if message:
                    LOGGER.info(
                        "Sending fatal diagnostic '%s' with message '%s'",
                        ex.get_diagnostic().uri,
                        message,
                    )
                else:
                    LOGGER.info(
                        "Sending fatal diagnostic '%s'", ex.get_diagnostic().uri
                    )
                LOGGER.debug("Fatal diagnostic was caused by this exception: %s", ex)

            # reset response buffer
            try:
                diagnostics = req.diagnostics
                if not diagnostics:
                    diagnostics = list()
                diagnostics.append(ex.get_diagnostic())

                buf = io.StringIO()
                out = self._create_XML_builder(
                    buf, SRURecordXmlEscaping.XML, False, req.get_indent_response()
                )
                ns = SRUNamespaces.get_namespaces(
                    req.get_version(), self.config.legacy_namespace_mode
                )
                self._write_fatal_error(out, ns, req, diagnostics)
                response.set_data(buf.getvalue())
            except Exception as ex:
                LOGGER.error("An exception occurred while in error state", ex)

    # TODO: temporary skip output buffering
    TEMP_OUTPUT_BUFFERING = False

    def explain(self, request: SRURequestImpl, response: Response):
        LOGGER.info("explain")

        # commence explain ...
        result = self.search_engine.explain(self.config, request, request)

        try:
            ns: SRUNamespaces = SRUNamespaces.get_namespaces(
                request.get_version(), self.config.legacy_namespace_mode
            )

            # send results
            buf = io.StringIO()
            out = self._create_XML_builder(
                buf,
                request.get_record_xml_escaping(),
                SRUServer.TEMP_OUTPUT_BUFFERING,
                request.get_indent_response(),
            )

            self._begin_response_with_request(out, ns, request)

            # write the explain record
            self._write_explain_record(out, ns, request)

            if self.config.echo_requests:
                self._write_echoed_explain_request(out, ns, request)

            # diagnostics
            self._write_diagnostics(out, ns, ns.response_NS, request.diagnostics)

            # extraResponseData
            if result:
                if result.has_extra_response_data:
                    with out.element("extraResponseData", ns.response_NS):
                        result.write_extra_response_data(out)

            self._end_response_with_request(out, ns, request)

            response.set_data(buf.getvalue())
        finally:
            if result:
                result.close()

    def scan(self, request: SRURequestImpl, response: Response):
        LOGGER.info("scan: scanClause = '%s'", request.get_scan_clause_raw())

        # commence scan ...
        result = self.search_engine.scan(self.config, request, request)
        if not result:
            raise SRUException(
                SRUDiagnostics.UNSUPPORTED_OPERATION,
                message="The 'scan' operation is not supported by this endpoint.",
            )

        try:
            ns: SRUNamespaces = SRUNamespaces.get_namespaces(
                request.get_version(), self.config.legacy_namespace_mode
            )

            # FIXME: re-check, if while scan response needs to be put
            # in scan namespace for SRU 2.0!

            # send results
            buf = io.StringIO()
            out = self._create_XML_builder(
                buf,
                request.get_record_xml_escaping(),
                SRUServer.TEMP_OUTPUT_BUFFERING,
                request.get_indent_response(),
            )

            self._begin_response_with_request(out, ns, request)

            try:
                # a scan result without a list of terms is a valid response;
                # make sure, to produce the correct output and omit in that
                # case the <terms> ...

                wrote_terms = False
                while result.next_term():
                    if not wrote_terms:
                        if ns.response_NS != ns.scan_NS:
                            out.startPrefixMapping(ns.scan_prefix, ns.scan_NS)
                            out.startElementNS((ns.scan_NS, "terms"))
                        wrote_terms = True

                    with out.element("term", ns.scan_NS):
                        with out.element("value", ns.scan_NS):
                            out.characters(result.get_value())

                        if result.get_number_of_records() > -1:
                            with out.element("numberOfRecords", ns.scan_NS):
                                out.characters(str(result.get_number_of_records()))

                        if result.get_display_term():
                            with out.element("displayTerm", ns.scan_NS):
                                out.characters(str(result.get_display_term()))

                        if result.get_WhereInList():
                            with out.element("whereInList", ns.scan_NS):
                                # NOTE: here it is not None
                                out.characters(result.get_WhereInList().lower())  # type: ignore

                        if result.has_extra_term_data():
                            with out.element("extraTermData", ns.scan_NS):
                                result.write_extra_term_data(out)

                if wrote_terms:
                    out.endElementNS((ns.scan_NS, "terms"))

            except StopIteration as ex:
                raise SRUException(
                    SRUDiagnostics.GENERAL_SYSTEM_ERROR,
                    message="An internal error occurred while serializing scan results.",
                ) from ex

            # echoedScanRequest
            if self.config.echo_requests:
                self._write_echoed_scan_request(
                    out, ns, request, request.get_scan_clause()
                )

            # diagnostics
            self._write_diagnostics(out, ns, ns.scan_NS, request.diagnostics)

            # extraResponseData
            if result.has_extra_response_data:
                with out.element("extraResponseData", ns.response_NS):
                    result.write_extra_response_data(out)

            self._end_response_with_request(out, ns, request)

            response.set_data(buf.getvalue())
        finally:
            if result:
                result.close()

    def search(self, request: SRURequestImpl, response: Response):
        LOGGER.info(
            "searchRetrieve: query = '%s', startRecord = %s, "
            "maximumRecords = %s, recordSchema = %s, resultSetTTL = %s",
            request.get_query_raw(),
            request.get_start_record(),
            request.get_maximum_records(),
            request.get_record_schema_identifier(),
            request.get_resultSet_TTL(),
        )

        # commence search ...
        result = self.search_engine.search(self.config, request, request)
        if not result:
            raise SRUException(
                SRUDiagnostics.GENERAL_SYSTEM_ERROR,
                message="SRUSearchEngine implementation returned invalid result (null).",
            )

        # check, of startRecord position is greater than total record set
        if (
            result.get_total_record_count() >= 0
            and request.get_start_record() > 1
            and request.get_start_record() > result.get_total_record_count()
        ):
            raise SRUException(SRUDiagnostics.FIRST_RECORD_POSITION_OUT_OF_RANGE)

        try:
            ns: SRUNamespaces = SRUNamespaces.get_namespaces(
                request.get_version(), self.config.legacy_namespace_mode
            )

            # send results
            buf = io.StringIO()
            out = self._create_XML_builder(
                buf,
                request.get_record_xml_escaping(),
                SRUServer.TEMP_OUTPUT_BUFFERING,
                request.get_indent_response(),
            )

            self._begin_response_with_request(out, ns, request)

            # numberOfRecords
            with out.element("numberOfRecords", ns.response_NS):
                out.characters(str(result.get_total_record_count()))

            # resultSetId
            if result.get_resultSet_id():
                with out.element("resultSetId", ns.response_NS):
                    out.characters(result.get_resultSet_id())

            # resultSetIdleTime (SRU 1.1 and SRU 1.2)
            if (
                not request.is_version(SRUVersion.VERSION_2_0)
                and result.get_resultSet_TTL() >= 0
            ):
                with out.element("resultSetIdleTime", ns.response_NS):
                    out.characters(str(result.get_resultSet_TTL()))

            position = (
                request.get_start_record() if request.get_start_record() > 0 else 1
            )

            if result.get_record_count() > 0:
                max_position_offset = (
                    (position + request.get_maximum_records() - 1)
                    if request.get_maximum_records() != -1
                    else -1
                )

                try:
                    out.startElementNS((ns.response_NS, "records"))

                    while result.next_record():
                        # Sanity check: do not return more then the maximum
                        # requested records. If the search engine
                        # implementation does not honor limit truncate the
                        # result set.
                        if max_position_offset != -1 and position > max_position_offset:
                            LOGGER.error(
                                "SRUSearchEngine implementation did not honor limit "
                                "for the amount of requsted records. Result set truncated!"
                            )
                            break

                        out.startElementNS((ns.response_NS, "record"))

                        # We need to output either the record or a surrogate
                        # diagnostic. In case of the latter, we need to output
                        # the appropriate record schema ...
                        diagnostic = result.get_surrogate_diagnostic()
                        with out.element("recordSchema", ns.response_NS):
                            if not diagnostic:
                                out.characters(result.get_record_schema_identifier())
                            else:
                                out.characters(SRU_DIAGNOSTIC_RECORD_SCHEMA)

                        # recordPacking (SRU 2.0). Only serialize, if it was in
                        # request.
                        # XXX: not sure, how to support 'unpacked' record
                        # packing anyways :/
                        if (
                            request.is_version(SRUVersion.VERSION_2_0)
                            and request.get_record_packing_raw()
                        ):
                            self._write_record_packing(
                                out, ns, request.get_record_packing()
                            )

                        # recordXMLEscaping (SRU 2.0) or
                        # recordPacking (SRU 1.1 and 1.2)
                        self._write_record_xml_escaping(out, ns, request)

                        # Output either record data or surrogate diagnostic ...
                        with out.element("recordData", ns.response_NS), out.record():
                            if diagnostic is None:
                                result.write_record(out)
                            else:
                                # write a surrogate diagnostic
                                self._write_diagnostic(out, ns, diagnostic, True)

                        # recordIdentifier is version 1.2+ only
                        if request.is_version_between(
                            SRUVersion.VERSION_1_2, SRUVersion.VERSION_2_0
                        ):
                            identifier = result.get_record_identifier()
                            if identifier:
                                with out.element("recordIdentifier", ns.response_NS):
                                    out.characters(identifier)

                        with out.element("recordPosition", ns.response_NS):
                            out.characters(str(position))

                        if result.has_extra_record_data:
                            with out.element("extraRecordData", ns.response_NS):
                                result.write_extra_record_data(out)

                        out.endElementNS((ns.response_NS, "record"))

                        position += 1

                    out.endElementNS((ns.response_NS, "records"))
                except StopIteration as ex:
                    raise SRUException(
                        SRUDiagnostics.GENERAL_SYSTEM_ERROR,
                        message="An internal error occurred while serializing search result set.",
                    ) from ex

            # nextRecordPosition
            if position <= result.get_total_record_count():
                with out.element("nextRecordPosition", ns.response_NS):
                    out.characters(str(position))

            # echoedSearchRetrieveRequest
            if self.config.echo_requests:
                # TODO: need to check, query should not be None
                self._write_echoed_searchRetrieve_request(
                    out, ns, request, request.get_query()  # type: ignore
                )

            # diagnostics
            self._write_diagnostics(out, ns, ns.response_NS, request.diagnostics)

            # extraResponseData
            if result.has_extra_response_data:
                with out.element("extraResponseData", ns.response_NS):
                    result.write_extra_response_data(out)

            # SRU 2.0 stuff ...
            if request.is_version(SRUVersion.VERSION_2_0):
                # resultSetTTL
                if result.get_resultSet_TTL() >= 0:
                    with out.element("resultSetTTL", ns.response_NS):
                        out.characters(str(result.get_resultSet_TTL()))

                # resultCountPrecision
                precision = result.get_result_count_precision()
                if precision:
                    with out.element("resultCountPrecision", ns.response_NS):
                        prefix = "info:srw/vocabulary/resultCountPrecision/1/"
                        out.characters(f"{prefix}{precision.lower()}")

                # facetedResults
                # NOTE: NOT YET SUPPORTED

                # searchResultAnalysis
                # NOTE: NOT YET SUPPORTED

            self._end_response_with_request(out, ns, request)

            response.set_data(buf.getvalue())
        finally:
            if result:
                result.close()

    # ----------------------------------------------------

    def _begin_response(
        self,
        out: SRUXMLStreamWriter,
        ns: SRUNamespaces,
        operation: SRUOperation,
        version: SRUVersion,
        stylesheet: Optional[str],
    ):
        out.startDocument()

        if stylesheet:
            out.processingInstruction(
                "xml-stylesheet", f'type="text/xsl" href="{stylesheet}"'
            )

        if operation == SRUOperation.EXPLAIN:
            out.startPrefixMapping(ns.response_prefix, ns.response_NS)
            out.startElementNS((ns.response_NS, "explainResponse"))
            self._write_version(out, ns.response_NS, version)
        elif operation == SRUOperation.SCAN:
            out.startPrefixMapping(ns.scan_prefix, ns.scan_NS)
            out.startElementNS((ns.scan_NS, "scanResponse"))
            self._write_version(out, ns.scan_NS, version)
        elif operation == SRUOperation.SEARCH_RETRIEVE:
            out.startPrefixMapping(ns.response_prefix, ns.response_NS)
            out.startElementNS((ns.response_NS, "searchRetrieveResponse"))
            self._write_version(out, ns.response_NS, version)

    def _begin_response_with_request(
        self, out: SRUXMLStreamWriter, ns: SRUNamespaces, request: SRURequest
    ):
        self._begin_response(
            out,
            ns,
            request.get_operation(),
            request.get_version(),
            request.get_stylesheet(),
        )

    def _end_response(
        self, out: SRUXMLStreamWriter, ns: SRUNamespaces, operation: SRUOperation
    ):
        if operation == SRUOperation.EXPLAIN:
            out.endElementNS((ns.response_NS, "explainResponse"))
        elif operation == SRUOperation.SCAN:
            out.endElementNS((ns.scan_NS, "scanResponse"))
        elif operation == SRUOperation.SEARCH_RETRIEVE:
            out.endElementNS((ns.response_NS, "searchRetrieveResponse"))

        out.endDocument()
        try:
            out.output_stream.flush()
            out.output_stream_raw.flush()

            # if we use buffers internally, then don't close them
            # otherwise we can't access the content anymore ...
            if not isinstance(out.output_stream, io.StringIO):
                out.output_stream.close()
            if not isinstance(out.output_stream_raw, io.StringIO):
                out.output_stream_raw.close()
        except Exception:
            pass

    def _end_response_with_request(
        self, out: SRUXMLStreamWriter, ns: SRUNamespaces, request: SRURequest
    ):
        self._end_response(out, ns, request.get_operation())

    # ----------------------------------------------------

    def _write_fatal_error(
        self,
        out: SRUXMLStreamWriter,
        ns: SRUNamespaces,
        request: SRURequestImpl,
        diagnostics: Optional[List[SRUDiagnostic]],
    ):
        # if operation is unknown, default to 'explain'
        operation = request.get_operation()
        if operation is None:
            operation = SRUOperation.EXPLAIN
        version = request.get_version()
        if version is None:
            version = self.config.default_version

        # write a response which conforms to the schema
        self._begin_response(out, ns, operation, version, None)
        if operation == SRUOperation.EXPLAIN:
            # 'explain' requires a complete explain record ...
            self._write_explain_record(out, ns, request)
            self._write_diagnostics(out, ns, ns.response_NS, diagnostics)
        elif operation == SRUOperation.SCAN:
            # 'scan' fortunately does not need any elements ...
            self._write_diagnostics(out, ns, ns.scan_NS, diagnostics)
        elif operation == SRUOperation.SEARCH_RETRIEVE:
            # 'searchRetrieve' needs numberOfRecords ..
            with out.element("numberOfRecords", ns.response_NS):
                out.characters("0")
            self._write_diagnostics(out, ns, ns.response_NS, diagnostics)

        self._end_response(out, ns, operation)

    def _write_diagnostics(
        self,
        out: SRUXMLStreamWriter,
        ns: SRUNamespaces,
        envelope_NS: str,
        diagnostics: Optional[List[SRUDiagnostic]],
    ):
        if not diagnostics:
            return

        out.startPrefixMapping(ns.diagnostic_prefix, ns.diagnostic_NS)
        with out.element("diagnostics", envelope_NS):
            for diagnostic in diagnostics:
                self._write_diagnostic(out, ns, diagnostic, False)

    def _write_diagnostic(
        self,
        out: SRUXMLStreamWriter,
        ns: SRUNamespaces,
        diagnostic: SRUDiagnostic,
        write_NS_decl: bool,
    ):
        if write_NS_decl:
            out.startPrefixMapping(ns.diagnostic_prefix, ns.diagnostic_NS)
        with out.element("diagnostic", ns.diagnostic_NS):
            with out.element("uri", ns.diagnostic_NS):
                out.characters(diagnostic.uri)
            if diagnostic.details:
                with out.element("details", ns.diagnostic_NS):
                    out.characters(diagnostic.details)
            if diagnostic.message:
                with out.element("message", ns.diagnostic_NS):
                    out.characters(diagnostic.message)

    # ----------------------------------------------------

    def _write_explain_record(
        self, out: SRUXMLStreamWriter, ns: SRUNamespaces, request: SRURequestImpl
    ):
        def _write_DatabaseInfo(info: Optional[DatabaseInfo]):
            if not info:
                return

            with out.element("databaseInfo", ns.explain_NS):
                self._write_LocalizedString(out, ns, "title", info.title)
                self._write_LocalizedString(out, ns, "description", info.description)

                self._write_LocalizedString(out, ns, "author", info.author)
                self._write_LocalizedString(out, ns, "extent", info.extent)
                self._write_LocalizedString(out, ns, "history", info.history)
                self._write_LocalizedString(out, ns, "langUsage", info.langUsage)
                self._write_LocalizedString(out, ns, "restrictions", info.restrictions)
                self._write_LocalizedString(out, ns, "subjects", info.subjects)
                self._write_LocalizedString(out, ns, "links", info.links)
                self._write_LocalizedString(
                    out, ns, "implementation", info.implementation
                )

        def _write_IndexInfo(info: Optional[IndexInfo]):
            if not info:
                return

            with out.element("indexInfo", ns.explain_NS):
                _write_IndexInfo_Sets(info.sets)
                _write_IndexInfo_Indexes(info.indexes)

        def _write_IndexInfo_Sets(sets: Optional[List[IndexInfo.Set]]):
            if not sets:
                return

            for set in sets:
                with out.element(
                    "set",
                    ns.explain_NS,
                    attrs={"identifier": set.identifier, "name": set.name},
                ):
                    self._write_LocalizedString(out, ns, "title", set.title)

        def _write_IndexInfo_Indexes(indexes: Optional[List[IndexInfo.Index]]):
            if not indexes:
                return

            for index in indexes:
                with out.element(
                    "index",
                    ns.explain_NS,
                    attrs={
                        "search": "true" if index.can_search else "false",
                        "scan": "true" if index.can_scan else "false",
                        "sort": "true" if index.can_sort else "false",
                    },
                ):
                    self._write_LocalizedString(out, ns, "title", index.title)

                    _write_IndexInfo_Index_Maps(index.maps)

        def _write_IndexInfo_Index_Maps(maps: Optional[List[IndexInfo.Index.Map]]):
            if not maps:
                return

            for map in maps:
                attrs: Dict[str, str] = dict()
                if map.primary:
                    attrs.update(primary="true")
                with out.element("map", ns.explain_NS, attrs=attrs):
                    with out.element(
                        "name",
                        ns.explain_NS,
                        attrs={"set": map.set},
                    ):
                        out.characters(map.name)

        def _write_SchemaInfos(infos: Optional[List[SchemaInfo]]):
            if not infos:
                return

            with out.element("schemaInfo", ns.explain_NS):
                for schema in infos:
                    attrs = {
                        "identifier": schema.identifier,
                        "name": schema.name,
                    }
                    # default is "false", so only add attribute if set to true
                    if schema.sort:
                        attrs.update({"sort": "true"})
                    # default is "true", so only add attribute if set to false
                    if not schema.retrieve:
                        attrs.update({"retrieve": "false"})

                    with out.element("schema", ns.explain_NS, attrs=attrs):
                        self._write_LocalizedString(out, ns, "title", schema.title)

        # ----------------------------

        with out.element("record", ns.response_NS):
            with out.element("recordSchema", ns.response_NS):
                out.characters(ns.explain_NS)

            # recordPacking (SRU 2.0)
            # Only serialize, if it was in request.
            # XXX: not sure, if this makes sense for explain
            if (
                request.is_version(SRUVersion.VERSION_2_0)
                and request.get_record_packing_raw()
            ):
                self._write_record_packing(out, ns, request.get_record_packing())

            # recordXMLEscaping (SRU 2.0) or recordPacking (SRU 1.1 and 1.2)
            self._write_record_xml_escaping(out, ns, request)

            with out.element("recordData", ns.response_NS), out.record():
                # explain ...
                out.startPrefixMapping(ns.explain_prefix, ns.explain_NS)
                with out.element("explain", ns.explain_NS):
                    # explain/serverInfo
                    with out.element(
                        "serverInfo",
                        ns.explain_NS,
                        attrs={
                            "protocol": "SRU",
                            "version": self.config.default_version.version_string,
                            "transport": self.config.transport,
                        },
                    ):
                        with out.element("host", ns.explain_NS):
                            out.characters(self.config.host)
                        with out.element("port", ns.explain_NS):
                            out.characters(str(self.config.port))
                        with out.element("database", ns.explain_NS):
                            out.characters(self.config.database)

                    # explain/databaseInfo
                    _write_DatabaseInfo(self.config.database_info)
                    # explain/indexInfo
                    _write_IndexInfo(self.config.index_info)
                    # explain/schemaInfo
                    _write_SchemaInfos(self.config.schema_info)

                    # explain/configInfo
                    with out.element("configInfo", ns.explain_NS):
                        # numberOfRecords (default)
                        with out.element(
                            "default", ns.explain_NS, attrs={"type": "numberOfRecords"}
                        ):
                            out.characters(str(self.config.number_of_records))

                        # maximumRecords (setting)
                        with out.element(
                            "setting", ns.explain_NS, attrs={"type": "maximumRecords"}
                        ):
                            out.characters(str(self.config.maximum_records))

    def _write_echoed_explain_request(
        self, out: SRUXMLStreamWriter, ns: SRUNamespaces, request: SRURequestImpl
    ):
        # echoedSearchRetrieveRequest ?
        with out.element("echoedExplainRequest", ns.response_NS):
            # echoedExplainRequest/version
            if request.get_version_raw() is not None:
                # NOTE: version is not None
                self._write_version(out, ns.response_NS, request.get_version_raw())  # type: ignore

            # echoedExplainRequest/recordXmlEscpaing / recordPacking
            if request.get_record_packing_raw():
                self._write_record_xml_escaping(out, ns, request)

            # echoedExplainRequest/stylesheet
            if request.stylesheet:
                with out.element("stylesheet", ns.response_NS):
                    out.characters(request.get_stylesheet())

    def _write_echoed_scan_request(
        self,
        out: SRUXMLStreamWriter,
        ns: SRUNamespaces,
        request: SRURequestImpl,
        query: Optional[cql.CQLQuery],
    ):
        # echoedScanRequest
        with out.element("echoedScanRequest", ns.response_NS):
            # echoedScanRequest/version
            if request.get_version_raw() is not None:
                # NOTE: version is not None
                self._write_version(out, ns.response_NS, request.get_version_raw())  # type: ignore

            # echoedScanRequest/scanClause
            with out.element("scanClause", ns.response_NS):
                out.characters(request.get_scan_clause_raw())

            # echoedScanRequest/xScanClause
            out.startPrefixMapping(None, ns.XCQL_NS)
            with out.element("xScanClause", ns.response_NS):
                # TODO: can this be None? it should not, need to test
                out.writeXCQL(query, False)  # type: ignore

            # echoedScanRequest/responsePosition
            if request.get_response_position() != -1:
                with out.element("responsePosition", ns.response_NS):
                    out.characters(str(request.get_response_position()))

            # echoedScanRequest/maximumTerms
            if request.get_maximum_terms() != -1:
                with out.element("maximumTerms", ns.response_NS):
                    out.characters(str(request.get_maximum_terms()))

            # echoedScanRequest/stylesheet
            if request.get_stylesheet():
                with out.element("stylesheet", ns.response_NS):
                    out.characters(request.get_stylesheet())

    def _write_echoed_searchRetrieve_request(
        self,
        out: SRUXMLStreamWriter,
        ns: SRUNamespaces,
        request: SRURequestImpl,
        query: SRUQuery[Any],
    ):
        # echoedSearchRetrieveRequest
        with out.element("echoedSearchRetrieveRequest", ns.response_NS):
            # echoedSearchRetrieveRequest/version
            if request.get_version_raw() is not None:
                # NOTE: version is not None
                self._write_version(out, ns.response_NS, request.get_version_raw())  # type: ignore

            # XXX: unclear, if <query> should only be echoed if queryType is CQL!?
            if SRUQueryType.CQL == query.query_type:
                # echoedSearchRetrieveRequest/query
                with out.element("query", ns.response_NS):
                    out.characters(query.raw_query)

                # echoedSearchRetrieveRequest/xQuery
                out.startPrefixMapping(None, ns.XCQL_NS)
                with out.element("xQuery", ns.response_NS):
                    out.writeXCQL(query.parsed_query, True)

            # echoedSearchRetrieveRequest/startRecord
            if request.get_start_record() > 0:
                with out.element("startRecord", ns.response_NS):
                    out.characters(str(request.get_start_record()))

            # echoedSearchRetrieveRequest/maximumRecords
            if request.get_maximum_records_raw() > 0:
                with out.element("maximumRecords", ns.response_NS):
                    out.characters(str(request.get_maximum_records_raw()))

            # (SRU 2.0) echoedSearchRetrieveRequest/recordPacking
            if (
                request.is_version(SRUVersion.VERSION_2_0)
                and request.get_record_packing_raw()
            ):
                with out.element("recordPacking", ns.response_NS):
                    out.characters(request.get_record_packing_raw())

            # echoedSearchRetrieveRequest/recordXmlEscaping / recordPacking
            if request.get_record_xml_escaping_raw():
                tag = (
                    "recordXMLEscaping"
                    if request.is_version(SRUVersion.VERSION_2_0)
                    else "recordPacking"
                )
                with out.element(tag, ns.response_NS):
                    out.characters(request.get_record_xml_escaping_raw())

            # echoedSearchRetrieveRequest/recordSchema
            if request.get_record_schema_identifier_raw():
                with out.element("recordSchema", ns.response_NS):
                    out.characters(request.get_record_schema_identifier_raw())

            # echoedSearchRetrieveRequest/recordXPath (1.1)
            if (
                request.is_version(SRUVersion.VERSION_1_1)
                and request.get_record_xpath()
            ):
                with out.element("recordXPath", ns.response_NS):
                    out.characters(request.get_record_xpath())

            # echoedSearchRetrieveRequest/resultSetTTL
            if request.get_resultSet_TTL() > 0:
                with out.element("resultSetTTL", ns.response_NS):
                    out.characters(str(request.get_resultSet_TTL()))

            # echoedSearchRetrieveRequest/sortKeys
            if request.is_version(SRUVersion.VERSION_1_1) and request.get_sortKeys():
                with out.element("sortKeys", ns.response_NS):
                    out.characters(request.get_sortKeys())

            # echoedSearchRetrieveRequest/xsortKeys

            # echoedSearchRetrieveRequest/stylesheet
            if request.get_stylesheet():
                with out.element("stylesheet", ns.response_NS):
                    out.characters(request.get_stylesheet())

            # echoedSearchRetrieveRequest/renderedBy
            if request.is_version(SRUVersion.VERSION_2_0) and request.get_renderBy():
                with out.element("renderedBy", ns.response_NS):
                    out.characters(request.get_renderBy().lower())  # type: ignore

            # echoedSearchRetrieveRequest/extraRequestParameter
            # FIXME: NOT YET IMPLEMENTED

            # echoedSearchRetrieveRequest/httpAccept
            # NOTE: broken in java version? uses renderedBy
            if (
                request.is_version(SRUVersion.VERSION_2_0)
                and request.get_http_accept_raw()
            ):
                with out.element("httpAccept", ns.response_NS):
                    out.characters(request.get_http_accept_raw())

            # echoedSearchRetrieveRequest/responseType
            if (
                request.is_version(SRUVersion.VERSION_2_0)
                and request.get_response_type()
            ):
                with out.element("responseType", ns.response_NS):
                    out.characters(request.get_response_type())

    # ----------------------------------------------------

    def _write_version(
        self, out: SRUXMLStreamWriter, envelope_NS: str, version: SRUVersion
    ):
        with out.element("version", envelope_NS):
            out.characters(version.version_string)

    def _write_record_xml_escaping(
        self, out: SRUXMLStreamWriter, ns: SRUNamespaces, request: SRURequest
    ):
        tag = (
            "recordXMLEscaping"
            if request.is_version(SRUVersion.VERSION_2_0)
            else "recordPacking"
        )
        with out.element(tag, ns.response_NS):
            out.characters(request.get_record_xml_escaping().lower())

    def _write_record_packing(
        self,
        out: SRUXMLStreamWriter,
        ns: SRUNamespaces,
        record_packing: SRURecordPacking,
    ):
        with out.element("recordPacking", ns.response_NS):
            out.characters(record_packing.lower())

    def _write_LocalizedString(
        self,
        out: SRUXMLStreamWriter,
        ns: SRUNamespaces,
        name: str,
        items: Optional[List[LocalizedString]],
    ):
        if not items:
            return
        for item in items:
            attrs: Dict[str, str] = dict()
            if item.lang:
                attrs.update(lang=item.lang)
            if item.primary:
                attrs.update(primary="true")
            with out.element(name, ns.explain_NS, attrs=attrs):
                out.characters(item.value)

    # ----------------------------------------------------

    def _create_XML_builder(
        self,
        output_stream: io.TextIOBase,
        record_packing: SRURecordXmlEscaping,
        skip_flush: bool,
        indent: int,
    ):
        try:
            if skip_flush:
                """
                Add a BufferedWriter(?) to delay flush() as long as possible.
                Doing so, enabled us to send an appropriate SRU diagnostic
                in case an error occurs during the serialization of the response.
                Of course, if an error occurs when the server response buffer
                already had been flushed, because it was to large, we cannot
                fail gracefully and we will produce ill-formed XML output.
                """
                # FIXME: do we have some...?
                # output_stream = io.BufferedWriter(
                #     output_stream, buffer_size=self.config.response_buffer_size
                # )
                # NOTE: might only make sense in async(io) use-case
                # otherwise always buffered in memory
                LOGGER.debug(
                    "Function for 'skip_flush' in '_create_XML_builder' not used."
                )

            return SRUXMLStreamWriter(
                output_stream, record_escaping=record_packing, indent=indent
            )
        except Exception as ex:
            raise SRUException(
                SRUDiagnostics.GENERAL_SYSTEM_ERROR,
                message="Error creating output stream.",
            ) from ex


# ---------------------------------------------------------------------------



from abc import ABC
from abc import abstractmethod
from typing import Any
from typing import Dict
from typing import Generic
from typing import List
from typing import Optional
from typing import Type
from typing import TypeVar

import cql

from .constants import SRUDiagnostics
from .constants import SRUParam
from .constants import SRUQueryType
from .constants import SRUVersion
from .diagnostic import SRUDiagnosticList
from .exception import SRUConfigException

_T = TypeVar("_T")


# ---------------------------------------------------------------------------


class SRUQuery(ABC, Generic[_T]):
    """Holder class for a parsed query to be returned from a
    `SRUQueryParser`."""

    def __init__(self, raw_query: str, parsed_query: _T):
        super().__init__()
        if raw_query is None:
            raise TypeError("raw_query is None")
        if parsed_query is None:
            raise TypeError("parsed_query is None")
        self._raw_query = raw_query
        self._parsed_query = parsed_query

    @property
    @abstractmethod
    def query_type(self) -> str:
        """Get the short name for supported query, e.g. "cql"."""

    @property
    def raw_query(self) -> str:
        """Get the original query as a string."""
        return self._raw_query

    @property
    def parsed_query(self) -> _T:
        """Get the parsed query as an abstract syntax tree."""
        return self._parsed_query


# ---------------------------------------------------------------------------


class SRUQueryParser(ABC, Generic[_T]):
    """Interface for implementing pluggable query parsers.

    Parameterized by 'abstract syntax tree (object) for parsed queries.'
    """

    @property
    @abstractmethod
    def query_type(self) -> str:
        """Get the short name for supported query, e.g. "cql"."""

    @abstractmethod
    def supports_version(self, version: Optional[SRUVersion]) -> bool:
        """Check if query is supported by a specific version of SRU/CQL."""

    @property
    def query_type_definition(self) -> Optional[str]:
        """The URI for the for the query type's definition."""
        return None

    @property
    @abstractmethod
    def query_parameter_names(self) -> List[str]:
        """Get the list of query parameters."""

    @abstractmethod
    def parse_query(
        self,
        version: SRUVersion,
        parameters: Dict[str, str],
        diagnostics: SRUDiagnosticList,
    ) -> Optional[SRUQuery[_T]]:
        """Parse a query into an abstract syntax tree.

        Args:
            version: the SRU version the request was made
            parameters: the request parameters containing the query
            diagnostics: a `SRUDiagnosticList` for storing fatal and
                non-fatal diagnostics

        Returns:
            the parsed query or ``None`` if the query could not be parsed
        """


# ---------------------------------------------------------------------------


class SRUQueryParserRegistry:
    """A registry to keep track of registered `SRUQueryParser` to be
    used by the `SRUServer`.

    See also:
        `SRUQueryParser`
    """

    def __init__(self, parsers: List[SRUQueryParser[Any]]):
        if parsers is None:
            raise TypeError("parsers is None")
        if not parsers:
            raise ValueError("parsers is empty")
        self.parsers = list(parsers)

    @property
    def query_parsers(self) -> List[SRUQueryParser[Any]]:
        """Get a list of all registered query parsers.

        Returns:
            List[SRUQueryParser[Any]]: a list of registered query
                parsers
        """
        return self.parsers

    def find_query_parser(self, query_type: str) -> Optional[SRUQueryParser[Any]]:
        """Find a query parser by query type.

        Args:
            query_type: the query type to search for

        Returns:
            SRUQueryParser[Any]: the matching `SRUQueryParser`
                instance or ``None`` if no matching parser was found.
        """
        if query_type is None:
            raise TypeError("query_type is None")
        return SRUQueryParserRegistry._find_parser(self.parsers, query_type)

    @staticmethod
    def _find_parser(
        parsers: List[SRUQueryParser[Any]], query_type: str
    ) -> Optional[SRUQueryParser[Any]]:
        for parser in parsers:
            if query_type == parser.query_type:
                return parser
        return None

    # ----------------------------------------------------

    class Builder:
        """Builder for creating `SRUQueryParserRegistry` instances."""

        def __init__(self, register_defaults: bool = True):
            """[Constructor]

            Args:
                register_defaults: if ``True``,
                    register SRU/CQL standard query parsers
                    (queryType **cql** and **searchTerms**),
                    otherwise do nothing. Defaults to True.
            """
            self.parsers: List[SRUQueryParser[Any]] = list()
            if register_defaults:
                self.register_defaults()

        def register_defaults(self) -> "SRUQueryParserRegistry.Builder":
            """Registers registers SRU/CQL standard query parsers
            (queryType **cql** and **searchTerms**)."""
            if not SRUQueryParserRegistry._find_parser(self.parsers, SRUQueryType.CQL):
                try:
                    self.register(CQLQueryParser())
                except SRUConfigException:
                    pass

            if not SRUQueryParserRegistry._find_parser(
                self.parsers, SRUQueryType.SEARCH_TERMS
            ):
                try:
                    self.register(SearchTermsQueryParser())
                except SRUConfigException:
                    pass

            return self

        def register(
            self, parser: SRUQueryParser[Any]
        ) -> "SRUQueryParserRegistry.Builder":
            """Register a new query parser

            Args:
                parser (SRUQueryParser[Any]): the query parser
                    instance to be registered

            Raises:
                `SRUConfigException`: if a query parser for the same
                    query type was already registered
            """
            if parser is None:
                raise TypeError("parser is None")
            if parser.query_type is None:
                raise TypeError("parser.query_type is None")

            # duplicate-save add ...
            if not SRUQueryParserRegistry._find_parser(self.parsers, parser.query_type):
                self.parsers.append(parser)
            else:
                raise SRUConfigException(
                    f"query parser for query_type '{parser.query_type}' is already registered"
                )

            return self

        def build(self) -> "SRUQueryParserRegistry":
            """Create a configured `SRUQueryParserRegistry` instance
            from this builder.

            Returns:
                SRUQueryParserRegistry: a `SRUQueryParserRegistry`
                    instance
            """
            return SRUQueryParserRegistry(self.parsers)


# ---------------------------------------------------------------------------


class SearchTermsQuery(SRUQuery[List[str]]):
    @property
    def query_type(self) -> str:
        return SRUQueryType.SEARCH_TERMS.value


class SearchTermsQueryParser(SRUQueryParser[List[str]]):
    @property
    def query_type(self) -> str:
        return SRUQueryType.SEARCH_TERMS

    @property
    def query_parameter_names(self) -> List[str]:
        return [SRUParam.QUERY.value]

    def supports_version(self, version: Optional[SRUVersion]) -> bool:
        if not version:
            raise TypeError("Argument version is invalid/None.")
        # java: version.compareTo(SRUVersion.VERSION_2_0) >= 0
        return version < SRUVersion.VERSION_2_0

    def parse_query(
        self,
        version: SRUVersion,
        parameters: Dict[str, str],
        diagnostics: SRUDiagnosticList,
    ) -> Optional[SRUQuery[List[str]]]:
        raw_query = parameters.get(SRUParam.QUERY)
        if raw_query is None:
            diagnostics.add_diagnostic(
                SRUDiagnostics.GENERAL_SYSTEM_ERROR,
                message="no query passed to query parser",
            )
            return None

        terms = raw_query.split()
        return SearchTermsQuery(raw_query, terms)


# ---------------------------------------------------------------------------


class CQLQuery(SRUQuery[cql.CQLQuery]):
    @property
    def query_type(self) -> str:
        return SRUQueryType.CQL.value


class CQLQueryParser(SRUQueryParser[cql.CQLQuery]):
    """Default query parser to parse CQL."""

    @property
    def query_type(self) -> str:
        return SRUQueryType.CQL

    @property
    def query_parameter_names(self) -> List[str]:
        return [SRUParam.QUERY.value]

    def supports_version(self, version: Optional[SRUVersion]) -> bool:
        if not version:
            raise TypeError("Argument version is invalid/None.")
        # CQL is supported by all SRU versions ...
        return True

    def parse_query(
        self,
        version: SRUVersion,
        parameters: Dict[str, str],
        diagnostics: SRUDiagnosticList,
    ) -> Optional[SRUQuery[cql.CQLQuery]]:
        raw_query = parameters.get(SRUParam.QUERY)
        if raw_query is None:
            diagnostics.add_diagnostic(
                SRUDiagnostics.GENERAL_SYSTEM_ERROR,
                message="no query passed to query parser",
            )
            return None

        # XXX: maybe query length against limit and return
        # "Too many characters in query" error?

        try:
            parser_cls: Type[cql.CQLParser] = cql.CQLParser12
            if version == SRUVersion.VERSION_1_1:
                parser_cls = cql.CQLParser11
            elif version in (SRUVersion.VERSION_1_2, SRUVersion.VERSION_2_0):
                parser_cls = cql.CQLParser12

            parser = parser_cls()
            parser.build()

            parsed_query = parser.parse(raw_query)
            return CQLQuery(raw_query, parsed_query)
        except cql.parser.CQLParserError:
            diagnostics.add_diagnostic(
                SRUDiagnostics.QUERY_SYNTAX_ERROR, message="error parsing query"
            )
        except Exception:
            diagnostics.add_diagnostic(
                SRUDiagnostics.QUERY_SYNTAX_ERROR, message="error parsing query"
            )

        return None


# ---------------------------------------------------------------------------



from abc import ABCMeta
from abc import abstractmethod
from dataclasses import dataclass
from typing import Optional

from . import constants

# ---------------------------------------------------------------------------


@dataclass(frozen=True)
class SRUDiagnostic:
    """Class to hold a SRU diagnostic.

    See also:
        * SRU Diagnostics: http://www.loc.gov/standards/sru/diagnostics/
        * SRU Diagnostics List: http://www.loc.gov/standards/sru/diagnostics/diagnosticsList.html
    """

    uri: str
    """Diagnostic's identifying URI."""

    details: Optional[str] = None
    """Supplementary information available, often in a format
    specified by the diagnostic or ``None``."""

    message: Optional[str] = None
    """Human readable message to display to the end user or ``None``."""

    def __post_init__(self):
        if not self.message or not self.message.strip():
            object.__setattr__(
                self, "message", self.get_default_error_message(self.uri)
            )

    @staticmethod
    def get_default_error_message(uri: str):
        diag = constants.SRUDiagnostics.get_by_uri(uri)
        if diag:
            return diag.description
        return None


# ---------------------------------------------------------------------------


class SRUDiagnosticList(metaclass=ABCMeta):
    """Container for non surrogate diagnostics for the request. The
    will be put in the ``diagnostics`` part of the response."""

    @abstractmethod
    def add_diagnostic(
        self, uri: str, details: Optional[str] = None, message: Optional[str] = None
    ) -> None:
        """Add a non surrogate diagnostic to the response.

        Args:
            uri: the diagnostic's identifying URI
            details: supplementary information available, often
                in a format specified by the diagnostic or ``None``
            message: human readable message to display to the
                end user or ``None``
        """


# ---------------------------------------------------------------------------



import io
from typing import Dict
from typing import Union

import pytest
from lxml import etree

import clarin.sru.server.config
from clarin.sru.server.config import SRUServerConfig
from clarin.sru.server.config import SRUServerConfigKey

ConfigParamDict = Dict[Union[SRUServerConfigKey, str], str]


# ---------------------------------------------------------------------------


@pytest.fixture(scope="function")
def config_params() -> ConfigParamDict:
    # minimal required params
    return {
        SRUServerConfigKey.SRU_TRANSPORT: "http",
        SRUServerConfigKey.SRU_HOST: "localhost",
        SRUServerConfigKey.SRU_PORT: "80",
        SRUServerConfigKey.SRU_DATABASE: "test",
    }


# ---------------------------------------------------------------------------


def test_load_config_file(server_config_content: bytes):
    buf = io.BytesIO(server_config_content)

    doc = SRUServerConfig.load_config_file(buf)
    assert doc is not None
    assert isinstance(doc, etree._ElementTree)


def test_build_infos(server_config_doc: etree._ElementTree):
    database_info = SRUServerConfig._build_DatabaseInfo(server_config_doc)
    assert isinstance(database_info, clarin.sru.server.config.DatabaseInfo)

    index_info = SRUServerConfig._build_IndexInfo(server_config_doc)
    assert isinstance(index_info, clarin.sru.server.config.IndexInfo)

    schema_infos = SRUServerConfig._build_SchemaInfo(server_config_doc)
    assert isinstance(schema_infos, list)
    assert all(
        isinstance(si, clarin.sru.server.config.SchemaInfo) for si in schema_infos
    )


def test_parse(server_config_content: bytes, config_params: ConfigParamDict):
    # "simulated" file stream
    buf = io.BytesIO(server_config_content)

    config = SRUServerConfig.parse(config_params, buf)
    assert isinstance(config, SRUServerConfig)


def test_parse_file(server_config_file: str, config_params: ConfigParamDict):
    config = SRUServerConfig.parse(config_params, server_config_file)
    assert isinstance(config, SRUServerConfig)


# ---------------------------------------------------------------------------



from typing import Optional

import cql
import pytest

from clarin.sru.constants import SRUParam
from clarin.sru.constants import SRUVersion
from clarin.sru.diagnostic import SRUDiagnosticList
from clarin.sru.queryparser import CQLQueryParser
from clarin.sru.queryparser import SearchTermsQuery
from clarin.sru.queryparser import SearchTermsQueryParser
from clarin.sru.queryparser import SRUQueryParser
from clarin.sru.queryparser import SRUQueryParserRegistry

# ---------------------------------------------------------------------------


def test_registry():
    with pytest.raises(TypeError):
        SRUQueryParserRegistry()

    with pytest.raises(TypeError):
        SRUQueryParserRegistry(None)

    with pytest.raises(ValueError):
        SRUQueryParserRegistry([])

    # NOTE: no type check; usually type hints (mypy etc) should check this?
    reg = SRUQueryParserRegistry(["parser1", "parser2"])
    assert ["parser1", "parser2"] == reg.parsers
    assert ["parser1", "parser2"] == reg.query_parsers

    # ----------------------------------------------------

    reg = SRUQueryParserRegistry.Builder(True).build()
    assert isinstance(
        reg.find_query_parser(SearchTermsQueryParser().query_type),
        SearchTermsQueryParser,
    )

    with pytest.raises(TypeError):
        reg.find_query_parser(None)


def test_registry_builder():
    bi = SRUQueryParserRegistry.Builder(False)
    assert not bi.parsers
    with pytest.raises(ValueError, match="parsers is empty"):
        bi.build()

    bi = SRUQueryParserRegistry.Builder(True)
    assert len(bi.parsers) == 2
    assert all(isinstance(p, SRUQueryParser) for p in bi.parsers)
    assert any(isinstance(p, SearchTermsQueryParser) for p in bi.parsers)
    assert any(isinstance(p, CQLQueryParser) for p in bi.parsers)
    reg = bi.build()
    assert len(reg.parsers) == 2
    assert len(reg.query_parsers) == 2
    assert all(isinstance(p, SRUQueryParser) for p in reg.parsers)
    assert any(isinstance(p, SearchTermsQueryParser) for p in reg.parsers)
    assert any(isinstance(p, CQLQueryParser) for p in reg.parsers)

    bi = SRUQueryParserRegistry.Builder(True)
    # only registers them once
    # NOTE: that subclassed queryparsers will also can not be registered
    bi.register_defaults()
    bi.register_defaults()
    bi.register_defaults()
    assert len(bi.parsers) == 2
    assert len(reg.query_parsers) == 2
    assert all(isinstance(p, SRUQueryParser) for p in bi.parsers)
    assert any(isinstance(p, SearchTermsQueryParser) for p in bi.parsers)
    assert any(isinstance(p, CQLQueryParser) for p in bi.parsers)

    bi = SRUQueryParserRegistry.Builder(False)
    bi.register_defaults()
    assert len(bi.parsers) == 2
    assert len(reg.query_parsers) == 2
    assert all(isinstance(p, SRUQueryParser) for p in bi.parsers)
    assert any(isinstance(p, SearchTermsQueryParser) for p in bi.parsers)
    assert any(isinstance(p, CQLQueryParser) for p in bi.parsers)


# ---------------------------------------------------------------------------


class MyTestDiagsList(SRUDiagnosticList):
    def __init__(self) -> None:
        self.diags = list()

    def add_diagnostic(
        self, uri: str, details: Optional[str] = None, message: Optional[str] = None
    ) -> None:
        self.diags.append((uri, details, message))


def test_SRUQuery():
    from clarin.sru.queryparser import CQLQuery
    from clarin.sru.queryparser import SRUQuery

    class TestSRUQuery(SRUQuery[str]):
        @property
        def query_type(self) -> str:
            return "test"

    with pytest.raises(TypeError):
        TestSRUQuery(None, None)
    with pytest.raises(TypeError):
        TestSRUQuery("abc", None)

    assert TestSRUQuery("abc", "def")

    with pytest.raises(TypeError):
        SearchTermsQuery(None, None)
    with pytest.raises(TypeError):
        SearchTermsQuery("abc", None)
    assert SearchTermsQuery("abc", "def")

    with pytest.raises(TypeError):
        CQLQuery(None, None)
    with pytest.raises(TypeError):
        CQLQuery("abc", None)
    # parameters not type checked
    assert CQLQuery("abc", "def")


def test_SearchTermsQueryParser():
    parser = SearchTermsQueryParser()
    assert not parser.supports_version(SRUVersion.VERSION_2_0)
    assert parser.supports_version(SRUVersion.VERSION_1_2)
    assert parser.query_type_definition is None

    diags = MyTestDiagsList()
    query = parser.parse_query(SRUVersion.VERSION_1_2, dict(), diags)
    assert diags.diags[0][2] == "no query passed to query parser"
    assert query is None

    diags = MyTestDiagsList()
    query = parser.parse_query(
        SRUVersion.VERSION_1_2, {SRUParam.QUERY: "test query"}, diags
    )
    assert not diags.diags
    assert query.raw_query == "test query"
    assert query.parsed_query == ["test", "query"]


def test_CQLQueryParser():
    parser = CQLQueryParser()
    assert parser.supports_version(SRUVersion.VERSION_1_1)
    assert parser.supports_version(SRUVersion.VERSION_2_0)
    assert parser.query_type_definition is None

    diags = MyTestDiagsList()
    query = parser.parse_query(SRUVersion.VERSION_2_0, dict(), diags)
    assert diags.diags[0][2] == "no query passed to query parser"
    assert query is None

    diags = MyTestDiagsList()
    query = parser.parse_query(
        SRUVersion.VERSION_2_0, {SRUParam.QUERY: "test query"}, diags
    )
    assert diags.diags[0][2] == "error parsing query"
    assert query is None

    diags = MyTestDiagsList()
    query = parser.parse_query(SRUVersion.VERSION_2_0, {SRUParam.QUERY: 1}, diags)
    assert diags.diags[0][2] == "error parsing query"
    assert query is None

    diags = MyTestDiagsList()
    query = parser.parse_query(
        SRUVersion.VERSION_2_0, {SRUParam.QUERY: '"test query"'}, diags
    )
    assert not diags.diags
    assert query.raw_query == '"test query"'
    assert isinstance(query.parsed_query.root, cql.parser.CQLSearchClause)
    assert query.parsed_query.root.term == "test query"


# ---------------------------------------------------------------------------



from abc import ABCMeta
from abc import abstractmethod
from typing import Dict
from typing import Optional

from werkzeug import Request

# ---------------------------------------------------------------------------


class SRUAuthenticationInfo(metaclass=ABCMeta):
    @property
    @abstractmethod
    def authentication_method(self) -> str:
        pass

    @property
    @abstractmethod
    def subject(self) -> str:
        pass


# ---------------------------------------------------------------------------


class SRUAuthenticationInfoProvider(metaclass=ABCMeta):
    @abstractmethod
    def get_AuthenticationInfo(
        self, request: Request
    ) -> Optional[SRUAuthenticationInfo]:
        # TODO: create wrapper around werkzeug.Request to allow more backends
        # TODO: or just use request headers as dict?
        pass


# ---------------------------------------------------------------------------


class SRUAuthenticationInfoProviderFactory(metaclass=ABCMeta):
    @abstractmethod
    def create_SRUAuthenticationInfoProvider(
        self, params: Dict[str, str]
    ) -> Optional[SRUAuthenticationInfoProvider]:
        """Create a authentication info provider."""
        pass


# ---------------------------------------------------------------------------
